package ru.dbotthepony.kstarbound.client.gl.vertex

import org.lwjgl.opengl.GL46
import ru.dbotthepony.kstarbound.client.gl.GLAttributeList
import ru.dbotthepony.kstarbound.client.gl.GLVertexBufferObject
import ru.dbotthepony.kstarbound.util.ByteBufferOutputStream

open class DirectVertexBuilder<T : DirectVertexBuilder<T>>(
	attributes: GLAttributeList,
	type: VertexType,
	val maxElements: Int,
) : AbstractVertexBuilder<DirectVertexBuilder<T>>(attributes, type) {
	val maxIndexCount = maxElements * type.indicies.size
	val maxVertexCount = maxElements * type.elements

	override val vertexMemory = ByteBufferOutputStream.directLE(maxVertexCount)
	override val elementMemory = ByteBufferOutputStream.directLE(maxIndexCount)

	override val elementIndexType: Int = when (maxIndexCount) {
		// api performance issue 102: glDrawElements uses element index type 'GL_UNSIGNED_BYTE' that is not optimal for the current hardware configuration; consider using 'GL_UNSIGNED_SHORT' instead
		// in 0 .. 255 -> GL_UNSIGNED_BYTE
		in 0 .. 65535 -> GL46.GL_UNSIGNED_SHORT
		else -> GL46.GL_UNSIGNED_INT
	}

	override fun ensureIndexCapacity() {
		if (vertexCount >= maxVertexCount) {
			throw IndexOutOfBoundsException("Vertex count overflow (can hold max of $maxElements elements, that's $maxVertexCount vertexes)")
		}
	}

	override fun resetMemory() {
		vertexMemory.position = 0
		elementMemory.position = 0
	}

	override fun doUpload(vbo: GLVertexBufferObject, ebo: GLVertexBufferObject, drawType: Int) {
		val vertexPos = vertexMemory.position
		val elementPos = elementMemory.position

		vertexMemory.position = 0
		elementMemory.position = 0

		vbo.bufferData(vertexMemory.buffer, drawType, length = vertexPos.toLong())
		ebo.bufferData(elementMemory.buffer, drawType, length = elementPos.toLong())

		vertexMemory.position = vertexPos
		elementMemory.position = elementPos
	}
}
