package ru.dbotthepony.kstarbound.util

import kotlin.reflect.KClass

class TwoDimensionalArray<T : Any>(clazz: KClass<T>, private val width: Int, private val height: Int) {
	private val memory: Array<T?> = java.lang.reflect.Array.newInstance(clazz.java, width * height) as Array<T?>

	operator fun get(x: Int, y: Int): T? {
		if (x !in 0 until width) {
			throw IndexOutOfBoundsException("X $x is out of bounds between 0 and $width")
		}

		if (y !in 0 until height) {
			throw IndexOutOfBoundsException("Y $y is out of bounds between 0 and $height")
		}

		return memory[x + y * width]
	}

	operator fun set(x: Int, y: Int, value: T): T? {
		if (x !in 0 until width) {
			throw IndexOutOfBoundsException("X $x is out of bounds between 0 and $width")
		}

		if (y !in 0 until height) {
			throw IndexOutOfBoundsException("Y $y is out of bounds between 0 and $height")
		}

		val old = memory[x + y * width]
		memory[x + y * width] = value
		return old
	}
}

inline fun <reified T : Any> TwoDimensionalArray(width: Int, height: Int) = TwoDimensionalArray(T::class, width, height)
